@@ -19,8 +19,8 @@ import math
 from typing import List
 
 from pyspark.sql import DataFrame
-from pyspark.sql.functions import explode, expr
 from pyspark.sql.functions import col
+from pyspark.sql.functions import explode, expr
 from pyspark.sql.types import StructType, StructField, IntegerType
 from shapely import wkt
 from shapely.wkt import loads
@@ -155,6 +155,10 @@ class TestPredicateJoin(TestBase):
         function_df = self.spark.sql("select ST_Distance(polygondf.countyshape, polygondf.countyshape) from polygondf")
         function_df.show()
 
+    def test_st_3ddistance(self):
+        function_df = self.spark.sql("select ST_3DDistance(ST_Point(0.0, 0.0, 5.0), ST_Point(1.0, 1.0, -6.0))")
+        assert function_df.count() == 1
+
     def test_st_transform(self):
         polygon_wkt_df = self.spark.read.format("csv"). \
             option("delimiter", "\t"). \
@@ -237,12 +241,95 @@ class TestPredicateJoin(TestBase):
         wkt_df = self.spark.sql("select ST_AsText(countyshape) as wkt from polygondf")
         assert polygon_df.take(1)[0]["countyshape"].wkt == loads(wkt_df.take(1)[0]["wkt"]).wkt
 
+
+    def test_st_astext_3d(self):
+        input_df = self.spark.createDataFrame([
+            ("Point(21 52 87)",),
+            ("Polygon((0 0 1, 0 1 1, 1 1 1, 1 0 1, 0 0 1))",),
+            ("Linestring(0 0 1, 1 1 2, 1 0 3)",),
+            ("MULTIPOINT ((10 40 66), (40 30 77), (20 20 88), (30 10 99))",),
+            ("MULTIPOLYGON (((30 20 11, 45 40 11, 10 40 11, 30 20 11)), ((15 5 11, 40 10 11, 10 20 11, 5 10 11, 15 5 11)))",),
+            ("MULTILINESTRING ((10 10 11, 20 20 11, 10 40 11), (40 40 11, 30 30 11, 40 20 11, 30 10 11))",),
+            ("MULTIPOLYGON (((40 40 11, 20 45 11, 45 30 11, 40 40 11)), ((20 35 11, 10 30 11, 10 10 11, 30 5 11, 45 20 11, 20 35 11), (30 20 11, 20 15 11, 20 25 11, 30 20 11)))",),
+            ("POLYGON((0 0 11, 0 5 11, 5 5 11, 5 0 11, 0 0 11), (1 1 11, 2 1 11, 2 2 11, 1 2 11, 1 1 11))",),
+        ], ["wkt"])
+
+        input_df.createOrReplaceTempView("input_wkt")
+        polygon_df = self.spark.sql("select ST_AsText(ST_GeomFromWkt(wkt)) as wkt from input_wkt")
+        assert polygon_df.count() == 8
+
+    def test_st_as_text_3d(self):
+        polygon_wkt_df = self.spark.read.format("csv"). \
+            option("delimiter", "\t"). \
+            option("header", "false"). \
+            load(mixed_wkt_geometry_input_location)
+
+        polygon_wkt_df.createOrReplaceTempView("polygontable")
+        polygon_df = self.spark.sql("select ST_GeomFromWKT(polygontable._c0) as countyshape from polygontable")
+        polygon_df.createOrReplaceTempView("polygondf")
+        wkt_df = self.spark.sql("select ST_AsText(countyshape) as wkt from polygondf")
+        assert polygon_df.take(1)[0]["countyshape"].wkt == loads(wkt_df.take(1)[0]["wkt"]).wkt
+
     def test_st_n_points(self):
         test = self.spark.sql("SELECT ST_NPoints(ST_GeomFromText('LINESTRING(77.29 29.07,77.42 29.26,77.27 29.31,77.29 29.07)'))")
 
     def test_st_geometry_type(self):
         test = self.spark.sql("SELECT ST_GeometryType(ST_GeomFromText('LINESTRING(77.29 29.07,77.42 29.26,77.27 29.31,77.29 29.07)'))")
 
+    def test_st_difference_right_overlaps_left(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))') as a,ST_GeomFromWKT('POLYGON ((0 -4, 4 -4, 4 4, 0 4, 0 -4))') as b")
+        test_table.createOrReplaceTempView("test_diff")
+        diff = self.spark.sql("select ST_Difference(a,b) from test_diff")
+        assert diff.take(1)[0][0].wkt == "POLYGON ((0 -3, -3 -3, -3 3, 0 3, 0 -3))"
+
+    def test_st_difference_right_not_overlaps_left(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))') as a,ST_GeomFromWKT('POLYGON ((5 -3, 7 -3, 7 -1, 5 -1, 5 -3))') as b")
+        test_table.createOrReplaceTempView("test_diff")
+        diff = self.spark.sql("select ST_Difference(a,b) from test_diff")
+        assert diff.take(1)[0][0].wkt == "POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))"
+
+    def test_st_difference_left_contains_right(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))') as a,ST_GeomFromWKT('POLYGON ((-1 -1, 1 -1, 1 1, -1 1, -1 -1))') as b")
+        test_table.createOrReplaceTempView("test_diff")
+        diff = self.spark.sql("select ST_Difference(a,b) from test_diff")
+        assert diff.take(1)[0][0].wkt == "POLYGON ((-3 -3, -3 3, 3 3, 3 -3, -3 -3), (-1 -1, 1 -1, 1 1, -1 1, -1 -1))"
+
+    def test_st_difference_right_not_overlaps_left(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-1 -1, 1 -1, 1 1, -1 1, -1 -1))') as a,ST_GeomFromWKT('POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))') as b")
+        test_table.createOrReplaceTempView("test_diff")
+        diff = self.spark.sql("select ST_Difference(a,b) from test_diff")
+        assert diff.take(1)[0][0].wkt == "POLYGON EMPTY"
+
+    def test_st_sym_difference_part_of_right_overlaps_left(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-1 -1, 1 -1, 1 1, -1 1, -1 -1))') as a,ST_GeomFromWKT('POLYGON ((0 -2, 2 -2, 2 0, 0 0, 0 -2))') as b")
+        test_table.createOrReplaceTempView("test_sym_diff")
+        diff = self.spark.sql("select ST_SymDifference(a,b) from test_sym_diff")
+        assert diff.take(1)[0][0].wkt == "MULTIPOLYGON (((0 -1, -1 -1, -1 1, 1 1, 1 0, 0 0, 0 -1)), ((0 -1, 1 -1, 1 0, 2 0, 2 -2, 0 -2, 0 -1)))"
+
+    def test_st_sym_difference_not_overlaps_left(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))') as a,ST_GeomFromWKT('POLYGON ((5 -3, 7 -3, 7 -1, 5 -1, 5 -3))') as b")
+        test_table.createOrReplaceTempView("test_sym_diff")
+        diff = self.spark.sql("select ST_SymDifference(a,b) from test_sym_diff")
+        assert diff.take(1)[0][0].wkt == "MULTIPOLYGON (((-3 -3, -3 3, 3 3, 3 -3, -3 -3)), ((5 -3, 5 -1, 7 -1, 7 -3, 5 -3)))"
+
+    def test_st_sym_difference_contains(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))') as a,ST_GeomFromWKT('POLYGON ((-1 -1, 1 -1, 1 1, -1 1, -1 -1))') as b")
+        test_table.createOrReplaceTempView("test_sym_diff")
+        diff = self.spark.sql("select ST_SymDifference(a,b) from test_sym_diff")
+        assert diff.take(1)[0][0].wkt == "POLYGON ((-3 -3, -3 3, 3 3, 3 -3, -3 -3), (-1 -1, 1 -1, 1 1, -1 1, -1 -1))"
+
+    def test_st_union_part_of_right_overlaps_left(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))') as a, ST_GeomFromWKT('POLYGON ((-2 1, 2 1, 2 4, -2 4, -2 1))') as b")
+        test_table.createOrReplaceTempView("test_union")
+        union = self.spark.sql("select ST_Union(a,b) from test_union")
+        assert union.take(1)[0][0].wkt == "POLYGON ((2 3, 3 3, 3 -3, -3 -3, -3 3, -2 3, -2 4, 2 4, 2 3))"
+
+    def test_st_union_not_overlaps_left(self):
+        test_table = self.spark.sql("select ST_GeomFromWKT('POLYGON ((-3 -3, 3 -3, 3 3, -3 3, -3 -3))') as a,ST_GeomFromWKT('POLYGON ((5 -3, 7 -3, 7 -1, 5 -1, 5 -3))') as b")
+        test_table.createOrReplaceTempView("test_union")
+        union = self.spark.sql("select ST_Union(a,b) from test_union")
+        assert union.take(1)[0][0].wkt == "MULTIPOLYGON (((-3 -3, -3 3, 3 3, 3 -3, -3 -3)), ((5 -3, 5 -1, 7 -1, 7 -3, 5 -3)))"
+
     def test_st_azimuth(self):
         sample_points = create_sample_points(20)
         sample_pair_points = [[el, sample_points[1]] for el in sample_points]
@@ -308,6 +395,30 @@ class TestPredicateJoin(TestBase):
 
         assert(not polygons.count())
 
+    def test_st_z(self):
+        point_df = self.spark.sql(
+            "select ST_GeomFromWKT('POINT Z (1.1 2.2 3.3)') as geom"
+        )
+        polygon_df = self.spark.sql(
+            "select ST_GeomFromWKT('POLYGON Z ((0 0 2, 0 1 2, 1 1 2, 1 0 2, 0 0 2))') as geom"
+        )
+        linestring_df = self.spark.sql(
+            "select ST_GeomFromWKT('LINESTRING Z (0 0 1, 0 1 2)') as geom"
+        )
+
+        points = point_df \
+            .selectExpr("ST_Z(geom)").collect()
+
+        polygons = polygon_df.selectExpr("ST_Z(geom) as z").filter("z IS NOT NULL")
+
+        linestrings = linestring_df.selectExpr("ST_Z(geom) as z").filter("z IS NOT NULL")
+
+        assert([point[0] for point in points] == [3.3])
+
+        assert(not linestrings.count())
+
+        assert(not polygons.count())
+
     def test_st_start_point(self):
 
         point_df = create_sample_points_df(self.spark, 5)
@@ -387,7 +498,7 @@ class TestPredicateJoin(TestBase):
 
     def test_st_exterior_ring(self):
         polygon_df = create_simple_polygons_df(self.spark, 5)
-        additional_wkt = "POLYGON((0 0 1, 1 1 1, 1 2 1, 1 1 1, 0 0 1))"
+        additional_wkt = "POLYGON((0 0, 1 1, 1 2, 1 1, 0 0))"
         additional_wkt_df = self.spark.createDataFrame([[wkt.loads(additional_wkt)]], self.geo_schema)
 
         polygons_df = polygon_df.union(additional_wkt_df)
@@ -733,6 +844,41 @@ class TestPredicateJoin(TestBase):
         for wkt, expected_wkt in geohash:
             assert wkt == expected_wkt
 
+    def test_st_collect_on_array_type(self):
+        # given
+        geometry_df = self.spark.createDataFrame([
+            [1, [loads("POLYGON((1 2,1 4,3 4,3 2,1 2))"), loads("POLYGON((0.5 0.5,5 0,5 5,0 5,0.5 0.5))")]],
+            [2, [loads("LINESTRING(1 2, 3 4)"), loads("LINESTRING(3 4, 4 5)")]],
+            [3, [loads("POINT(1 2)"), loads("POINT(-2 3)")]]
+        ]).toDF("id", "geom")
+
+        # when calculating st collect
+        geometry_df_collected = geometry_df.withColumn("collected", expr("ST_Collect(geom)"))
+
+        # then result should be as expected
+        assert(set([el[0] for el in  geometry_df_collected.selectExpr("ST_AsText(collected)").collect()]) == {
+            "MULTILINESTRING ((1 2, 3 4), (3 4, 4 5))",
+            "MULTIPOINT ((1 2), (-2 3))",
+            "MULTIPOLYGON (((1 2, 1 4, 3 4, 3 2, 1 2)), ((0.5 0.5, 5 0, 5 5, 0 5, 0.5 0.5)))"
+        })
+
+    def test_st_collect_on_multiple_columns(self):
+        # given geometry df with multiple geometry columns
+        geometry_df = self.spark.createDataFrame([
+            [1, loads("POLYGON((1 2,1 4,3 4,3 2,1 2))"), loads("POLYGON((0.5 0.5,5 0,5 5,0 5,0.5 0.5))")],
+            [2, loads("LINESTRING(1 2, 3 4)"), loads("LINESTRING(3 4, 4 5)")],
+            [3, loads("POINT(1 2)"), loads("POINT(-2 3)")]
+        ]).toDF("id", "geom_left", "geom_right")
+
+        # when calculating st collect on multiple columns
+        geometry_df_collected = geometry_df.withColumn("collected", expr("ST_Collect(geom_left, geom_right)"))
+
+        # then result should be calculated
+        assert(set([el[0] for el in  geometry_df_collected.selectExpr("ST_AsText(collected)").collect()]) == {
+            "MULTILINESTRING ((1 2, 3 4), (3 4, 4 5))",
+            "MULTIPOINT ((1 2), (-2 3))",
+            "MULTIPOLYGON (((1 2, 1 4, 3 4, 3 2, 1 2)), ((0.5 0.5, 5 0, 5 5, 0 5, 0.5 0.5)))"
+        })
 
     def calculate_st_is_ring(self, wkt):
         geometry_collected = self.__wkt_list_to_data_frame([wkt]). \
