@@ -21,6 +21,7 @@ package org.apache.sedona.sql.UDF
 import org.apache.spark.sql.catalyst.analysis.FunctionRegistry.FunctionBuilder
 import org.apache.spark.sql.expressions.{Aggregator, UserDefinedAggregateFunction}
 import org.apache.spark.sql.sedona_sql.expressions._
+import org.apache.spark.sql.sedona_sql.expressions.collect.ST_Collect
 import org.apache.spark.sql.sedona_sql.expressions.raster.{RS_AddBands, RS_Array, RS_Base64, RS_BitwiseAnd, RS_BitwiseOr, RS_Count, RS_DivideBands, RS_FetchRegion, RS_GetBand, RS_GreaterThan, RS_GreaterThanEqual, RS_HTML, RS_LessThan, RS_LessThanEqual, RS_LogicalDifference, RS_LogicalOver, RS_Mean, RS_Mode, RS_Modulo, RS_MultiplyBands, RS_MultiplyFactor, RS_Normalize, RS_NormalizedDifference, RS_SquareRoot, RS_SubtractBands}
 import org.locationtech.jts.geom.Geometry
 
@@ -40,6 +41,7 @@ object Catalog {
     ST_Intersects,
     ST_Within,
     ST_Distance,
+    ST_3DDistance,
     ST_ConvexHull,
     ST_NPoints,
     ST_Buffer,
@@ -49,6 +51,9 @@ object Catalog {
     ST_Centroid,
     ST_Transform,
     ST_Intersection,
+    ST_Difference,
+    ST_SymDifference,
+    ST_Union,
     ST_IsValid,
     ST_PrecisionReduce,
     ST_Equals,
@@ -70,6 +75,7 @@ object Catalog {
     ST_Azimuth,
     ST_X,
     ST_Y,
+    ST_Z,
     ST_StartPoint,
     ST_Boundary,
     ST_MinimumBoundingRadius,
@@ -93,7 +99,8 @@ object Catalog {
     ST_MakePolygon,
     ST_GeoHash,
     ST_GeomFromGeoHash,
-
+    ST_Collect,
+    ST_Multi,
     // Expression for rasters
     RS_NormalizedDifference,
     RS_Mean,
