@@ -27,6 +27,7 @@ import org.apache.spark.sql.catalyst.expressions.codegen.{CodegenContext, Codege
 import org.apache.spark.sql.catalyst.expressions.{BoundReference, Expression, Generator}
 import org.apache.spark.sql.catalyst.util.{ArrayData, GenericArrayData}
 import org.apache.spark.sql.sedona_sql.UDT.GeometryUDT
+import org.apache.spark.sql.sedona_sql.expressions.collect.Collect
 import org.apache.spark.sql.sedona_sql.expressions.geohash.{GeoHashDecoder, GeometryGeoHashEncoder, InvalidGeoHashException}
 import org.apache.spark.sql.sedona_sql.expressions.implicits._
 import org.apache.spark.sql.sedona_sql.expressions.subdivide.GeometrySubDivider
@@ -36,10 +37,11 @@ import org.geotools.geometry.jts.JTS
 import org.geotools.referencing.CRS
 import org.locationtech.jts.algorithm.MinimumBoundingCircle
 import org.locationtech.jts.geom.{PrecisionModel, _}
-import org.locationtech.jts.io.{ByteOrderValues, WKBWriter}
+import org.locationtech.jts.io.{ByteOrderValues, WKBWriter, WKTWriter}
 import org.locationtech.jts.linearref.LengthIndexedLine
 import org.locationtech.jts.operation.IsSimpleOp
 import org.locationtech.jts.operation.buffer.BufferParameters
+import org.locationtech.jts.operation.distance3d.Distance3DOp
 import org.locationtech.jts.operation.linemerge.LineMerger
 import org.locationtech.jts.operation.valid.IsValidOp
 import org.locationtech.jts.precision.GeometryPrecisionReducer
@@ -48,8 +50,6 @@ import org.opengis.referencing.operation.MathTransform
 import org.wololo.jts2geojson.GeoJSONWriter
 
 import java.nio.ByteOrder
-import java.util
-import scala.collection.mutable.ArrayBuffer
 import scala.util.{Failure, Success, Try}
 
 /**
@@ -58,21 +58,12 @@ import scala.util.{Failure, Success, Try}
   * @param inputExpressions This function takes two geometries and calculates the distance between two objects.
   */
 case class ST_Distance(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends BinaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 2)
 
-  override def nullable: Boolean = false
-
   override def toString: String = s" **${ST_Distance.getClass.getName}**  "
 
-  override def eval(inputRow: InternalRow): Any = {
-    val leftArray = inputExpressions(0).eval(inputRow).asInstanceOf[ArrayData]
-    val rightArray = inputExpressions(1).eval(inputRow).asInstanceOf[ArrayData]
-
-    val leftGeometry = GeometrySerializer.deserialize(leftArray)
-
-    val rightGeometry = GeometrySerializer.deserialize(rightArray)
-
+  override def nullSafeEval(leftGeometry: Geometry, rightGeometry: Geometry): Any = {
     leftGeometry.distance(rightGeometry)
   }
 
@@ -85,19 +76,35 @@ case class ST_Distance(inputExpressions: Seq[Expression])
   }
 }
 
+case class ST_3DDistance(inputExpressions: Seq[Expression])
+  extends BinaryGeometryExpression with CodegenFallback {
+  assert(inputExpressions.length == 2)
+
+  override def toString: String = s" **${ST_3DDistance.getClass.getName}**  "
+
+  override def nullSafeEval(leftGeometry: Geometry, rightGeometry: Geometry): Any = {
+    Distance3DOp.distance(leftGeometry, rightGeometry)
+  }
+
+  override def dataType = DoubleType
+
+  override def children: Seq[Expression] = inputExpressions
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
+}
+
 /**
   * Return the convex hull of a Geometry.
   *
   * @param inputExpressions
   */
 case class ST_ConvexHull(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
+  override def nullSafeEval(geometry: Geometry): Any = {
     new GenericArrayData(GeometrySerializer.serialize(geometry.convexHull()))
   }
 
@@ -116,16 +123,10 @@ case class ST_ConvexHull(inputExpressions: Seq[Expression])
   * @param inputExpressions
   */
 case class ST_NPoints(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
-  override def nullable: Boolean = false
+  extends UnaryGeometryExpression with CodegenFallback {
 
-  override def eval(input: InternalRow): Any = {
-    inputExpressions.length match {
-      case 1 =>
-        val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
-        geometry.getCoordinates.length
-      case _ => None
-    }
+  override def nullSafeEval(geometry: Geometry): Any = {
+    geometry.getCoordinates.length
   }
 
   override def dataType: DataType = IntegerType
@@ -146,16 +147,18 @@ case class ST_Buffer(inputExpressions: Seq[Expression])
   extends Expression with CodegenFallback {
   assert(inputExpressions.length == 2)
 
-  override def nullable: Boolean = false
+  override def nullable: Boolean = true
 
   override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions(0).eval(input).asInstanceOf[ArrayData])
     val buffer: Double = inputExpressions(1).eval(input) match {
       case a: Decimal => a.toDouble
       case a: Double => a
       case a: Int => a
     }
-    new GenericArrayData(GeometrySerializer.serialize(geometry.buffer(buffer)))
+    inputExpressions(0).toGeometry(input) match {
+      case geometry: Geometry => geometry.buffer(buffer).toGenericArrayData
+      case _ => null
+    }
   }
 
   override def dataType: DataType = GeometryUDT
@@ -174,13 +177,10 @@ case class ST_Buffer(inputExpressions: Seq[Expression])
   * @param inputExpressions
   */
 case class ST_Envelope(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
+  override def nullSafeEval(geometry: Geometry): Any = {
     new GenericArrayData(GeometrySerializer.serialize(geometry.getEnvelope()))
   }
 
@@ -199,13 +199,10 @@ case class ST_Envelope(inputExpressions: Seq[Expression])
   * @param inputExpressions
   */
 case class ST_Length(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
+  override def nullSafeEval(geometry: Geometry): Any = {
     geometry.getLength
   }
 
@@ -224,13 +221,10 @@ case class ST_Length(inputExpressions: Seq[Expression])
   * @param inputExpressions
   */
 case class ST_Area(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
+  override def nullSafeEval(geometry: Geometry): Any = {
     geometry.getArea
   }
 
@@ -249,13 +243,10 @@ case class ST_Area(inputExpressions: Seq[Expression])
   * @param inputExpressions
   */
 case class ST_Centroid(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions(0).eval(input).asInstanceOf[ArrayData])
+  override def nullSafeEval(geometry: Geometry): Any = {
     new GenericArrayData(GeometrySerializer.serialize(geometry.getCentroid()))
   }
 
@@ -277,22 +268,27 @@ case class ST_Transform(inputExpressions: Seq[Expression])
   extends Expression with CodegenFallback {
   assert(inputExpressions.length >= 3 && inputExpressions.length <= 4)
 
-  override def nullable: Boolean = false
+  override def nullable: Boolean = true
 
   override def eval(input: InternalRow): Any = {
-    val originalGeometry = GeometrySerializer.deserialize(inputExpressions(0).eval(input).asInstanceOf[ArrayData])
-    val sourceCRScode = CRS.decode(inputExpressions(1).eval(input).asInstanceOf[UTF8String].toString)
-    val targetCRScode = CRS.decode(inputExpressions(2).eval(input).asInstanceOf[UTF8String].toString)
+    val originalGeometry = inputExpressions(0).toGeometry(input)
+    val sourceCRS = inputExpressions(1).asString(input)
+    val targetCRS = inputExpressions(2).asString(input)
 
-    var transform: MathTransform = null
-    if (inputExpressions.length == 4) {
-      transform = CRS.findMathTransform(sourceCRScode, targetCRScode, inputExpressions(3).eval(input).asInstanceOf[Boolean])
+    (originalGeometry, sourceCRS, targetCRS) match {
+      case (originalGeometry: Geometry, sourceCRS: String, targetCRS: String) =>
+        val sourceCRScode = CRS.decode(sourceCRS)
+        val targetCRScode = CRS.decode(targetCRS)
+        var transform: MathTransform = null
+        if (inputExpressions.length == 4) {
+          transform = CRS.findMathTransform(sourceCRScode, targetCRScode, inputExpressions(3).eval(input).asInstanceOf[Boolean])
+        }
+        else {
+          transform = CRS.findMathTransform(sourceCRScode, targetCRScode, false)
+        }
+        JTS.transform(originalGeometry, transform).toGenericArrayData
+      case (_, _, _) => null
     }
-    else {
-      transform = CRS.findMathTransform(sourceCRScode, targetCRScode, false)
-    }
-    val geom = JTS.transform(originalGeometry, transform)
-    new GenericArrayData(GeometrySerializer.serialize(geom))
   }
 
   override def dataType: DataType = GeometryUDT
@@ -311,35 +307,30 @@ case class ST_Transform(inputExpressions: Seq[Expression])
   * @param inputExpressions
   */
 case class ST_Intersection(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends BinaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 2)
 
   lazy val GeometryFactory = new GeometryFactory()
   lazy val emptyPolygon = GeometryFactory.createPolygon(null, null)
 
-  override def nullable: Boolean = false
-
-  override def eval(inputRow: InternalRow): Any = {
-    val leftgeometry = GeometrySerializer.deserialize(inputExpressions(0).eval(inputRow).asInstanceOf[ArrayData])
-    val rightgeometry = GeometrySerializer.deserialize(inputExpressions(1).eval(inputRow).asInstanceOf[ArrayData])
-
-    val isIntersects = leftgeometry.intersects(rightgeometry)
-    lazy val isLeftContainsRight = leftgeometry.contains(rightgeometry)
-    lazy val isRightContainsLeft = rightgeometry.contains(leftgeometry)
+  override def nullSafeEval(leftGeometry: Geometry, rightGeometry: Geometry): Any = {
+    val isIntersects = leftGeometry.intersects(rightGeometry)
+    lazy val isLeftContainsRight = leftGeometry.contains(rightGeometry)
+    lazy val isRightContainsLeft = rightGeometry.contains(leftGeometry)
 
     if (!isIntersects) {
       return new GenericArrayData(GeometrySerializer.serialize(emptyPolygon))
     }
 
     if (isIntersects && isLeftContainsRight) {
-      return new GenericArrayData(GeometrySerializer.serialize(rightgeometry))
+      return new GenericArrayData(GeometrySerializer.serialize(rightGeometry))
     }
 
     if (isIntersects && isRightContainsLeft) {
-      return new GenericArrayData(GeometrySerializer.serialize(leftgeometry))
+      return new GenericArrayData(GeometrySerializer.serialize(leftGeometry))
     }
 
-    return new GenericArrayData(GeometrySerializer.serialize(leftgeometry.intersection(rightgeometry)))
+    new GenericArrayData(GeometrySerializer.serialize(leftGeometry.intersection(rightGeometry)))
   }
 
   override def dataType: DataType = GeometryUDT
@@ -408,19 +399,12 @@ case class ST_MakeValid(inputExpressions: Seq[Expression])
   * @param inputExpressions
   */
 case class ST_IsValid(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
 
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.eval(input).asInstanceOf[ArrayData]
-    if (geometry == null) {
-      null
-    } else {
-      val isvalidop = new IsValidOp(GeometrySerializer.deserialize(geometry))
-      isvalidop.isValid
-    }
+  override protected def nullSafeEval(geometry: Geometry): Any = {
+    new IsValidOp(geometry).isValid
   }
 
   override def dataType: DataType = BooleanType
@@ -438,15 +422,11 @@ case class ST_IsValid(inputExpressions: Seq[Expression])
   * @param inputExpressions
   */
 case class ST_IsSimple(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
-    val isSimpleop = new IsSimpleOp(geometry)
-    isSimpleop.isSimple
+  override protected def nullSafeEval(geometry: Geometry): Any = {
+    new IsSimpleOp(geometry).isSimple
   }
 
   override def dataType: DataType = BooleanType
@@ -470,18 +450,18 @@ case class ST_SimplifyPreserveTopology(inputExpressions: Seq[Expression])
   extends Expression with CodegenFallback {
   assert(inputExpressions.length == 2)
 
-  override def nullable: Boolean = false
+  override def nullable: Boolean = true
 
   override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions(0).eval(input).asInstanceOf[ArrayData])
     val distanceTolerance = inputExpressions(1).eval(input) match {
       case number: Decimal => number.toDouble
       case number: Double => number
       case number: Int => number.toDouble
     }
-    val simplifiedGeometry = TopologyPreservingSimplifier.simplify(geometry, distanceTolerance)
-
-    new GenericArrayData(GeometrySerializer.serialize(simplifiedGeometry))
+    inputExpressions(0).toGeometry(input) match {
+      case geometry: Geometry => TopologyPreservingSimplifier.simplify(geometry, distanceTolerance).toGenericArrayData
+      case _ => null
+    }
   }
 
   override def dataType: DataType = GeometryUDT
@@ -501,13 +481,16 @@ case class ST_SimplifyPreserveTopology(inputExpressions: Seq[Expression])
   */
 case class ST_PrecisionReduce(inputExpressions: Seq[Expression])
   extends Expression with CodegenFallback {
-  override def nullable: Boolean = false
+  override def nullable: Boolean = true
 
   override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions(0).eval(input).asInstanceOf[ArrayData])
     val precisionScale = inputExpressions(1).eval(input).asInstanceOf[Int]
-    val precisionReduce = new GeometryPrecisionReducer(new PrecisionModel(Math.pow(10, precisionScale)))
-    new GenericArrayData(GeometrySerializer.serialize(precisionReduce.reduce(geometry)))
+    inputExpressions(0).toGeometry(input) match {
+      case geometry: Geometry =>
+        val precisionReduce =new GeometryPrecisionReducer(new PrecisionModel(Math.pow(10, precisionScale)))
+        precisionReduce.reduce(geometry).toGenericArrayData
+      case _ => null
+    }
   }
 
   override def dataType: DataType = GeometryUDT
@@ -520,14 +503,12 @@ case class ST_PrecisionReduce(inputExpressions: Seq[Expression])
 }
 
 case class ST_AsText(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
-    UTF8String.fromString(geometry.toText)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
+    val writer = new WKTWriter(GeometrySerializer.getDimension(geometry))
+    UTF8String.fromString(writer.write(geometry))
   }
 
   override def dataType: DataType = StringType
@@ -540,12 +521,9 @@ case class ST_AsText(inputExpressions: Seq[Expression])
 }
 
 case class ST_AsGeoJSON(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  extends UnaryGeometryExpression with CodegenFallback {
 
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     val writer = new GeoJSONWriter()
     UTF8String.fromString(writer.write(geometry).toString)
   }
@@ -560,14 +538,11 @@ case class ST_AsGeoJSON(inputExpressions: Seq[Expression])
 }
 
 case class ST_AsBinary(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   inputExpressions.validateLength(1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
-    val dimensions = if (java.lang.Double.isNaN(geometry.getCoordinate.getZ)) 2 else 3
+  override protected def nullSafeEval(geometry: Geometry): Any = {
+    val dimensions = if (geometry.isEmpty() || java.lang.Double.isNaN(geometry.getCoordinate.getZ)) 2 else 3
     val endian = if (ByteOrder.nativeOrder() == ByteOrder.BIG_ENDIAN) ByteOrderValues.BIG_ENDIAN else ByteOrderValues.LITTLE_ENDIAN
     val writer = new WKBWriter(dimensions, endian)
     writer.write(geometry)
@@ -583,14 +558,11 @@ case class ST_AsBinary(inputExpressions: Seq[Expression])
 }
 
 case class ST_AsEWKB(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   inputExpressions.validateLength(1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
-    val dimensions = if (java.lang.Double.isNaN(geometry.getCoordinate.getZ)) 2 else 3
+  override protected def nullSafeEval(geometry: Geometry): Any = {
+    val dimensions = if (geometry.isEmpty() || java.lang.Double.isNaN(geometry.getCoordinate.getZ)) 2 else 3
     val endian = if (ByteOrder.nativeOrder() == ByteOrder.BIG_ENDIAN) ByteOrderValues.BIG_ENDIAN else ByteOrderValues.LITTLE_ENDIAN
     val writer = new WKBWriter(dimensions, endian, geometry.getSRID != 0)
     writer.write(geometry)
@@ -606,13 +578,10 @@ case class ST_AsEWKB(inputExpressions: Seq[Expression])
 }
 
 case class ST_SRID(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   inputExpressions.validateLength(1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry.getSRID
   }
 
@@ -629,13 +598,16 @@ case class ST_SetSRID(inputExpressions: Seq[Expression])
   extends Expression with CodegenFallback {
   inputExpressions.validateLength(2)
 
-  override def nullable: Boolean = false
+  override def nullable: Boolean = true
 
   override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
     val srid = inputExpressions(1).eval(input).asInstanceOf[Integer]
-    val factory = new GeometryFactory(geometry.getPrecisionModel, srid, geometry.getFactory.getCoordinateSequenceFactory)
-    new GenericArrayData(GeometrySerializer.serialize(factory.createGeometry(geometry)))
+    inputExpressions(0).toGeometry(input) match {
+      case geometry: Geometry =>
+        val factory = new GeometryFactory(geometry.getPrecisionModel, srid, geometry.getFactory.getCoordinateSequenceFactory)
+        factory.createGeometry(geometry).toGenericArrayData
+      case _ => null
+    }
   }
 
   override def dataType: DataType = GeometryUDT
@@ -648,13 +620,10 @@ case class ST_SetSRID(inputExpressions: Seq[Expression])
 }
 
 case class ST_GeometryType(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     UTF8String.fromString("ST_" + geometry.getGeometryType)
   }
 
@@ -675,18 +644,14 @@ case class ST_GeometryType(inputExpressions: Seq[Expression])
   * @param inputExpressions Geometry
   */
 case class ST_LineMerge(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
   // Definition of the Geometry Collection Empty
   lazy val GeometryFactory = new GeometryFactory()
   lazy val emptyGeometry = GeometryFactory.createGeometryCollection(null)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
-
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     val merger = new LineMerger()
 
     val output: Geometry = geometry match {
@@ -719,14 +684,12 @@ case class ST_LineMerge(inputExpressions: Seq[Expression])
 }
 
 case class ST_Azimuth(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends BinaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 2)
-  override def nullable: Boolean = false
 
-  override def eval(input: InternalRow): Any = {
-    val geometries = (inputExpressions(0).toGeometry(input), inputExpressions(1).toGeometry(input))
-    geometries match {
-       case (pointA: Point, pointB: Point) => calculateAzimuth(pointA, pointB)
+  override def nullSafeEval(leftGeometry: Geometry, rightGeometry: Geometry): Any = {
+    (leftGeometry, rightGeometry) match {
+      case (pointA: Point, pointB: Point) => calculateAzimuth(pointA, pointB)
     }
   }
 
@@ -747,14 +710,10 @@ case class ST_Azimuth(inputExpressions: Seq[Expression])
 }
 
 case class ST_X(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
-
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry match {
       case point: Point => point.getX
       case _ => null
@@ -772,14 +731,10 @@ case class ST_X(inputExpressions: Seq[Expression])
 
 
 case class ST_Y(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
-
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry match {
       case point: Point => point.getY
       case _ => null
@@ -795,14 +750,31 @@ case class ST_Y(inputExpressions: Seq[Expression])
   }
 }
 
-case class ST_StartPoint(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+case class ST_Z(inputExpressions: Seq[Expression])
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
+  override protected def nullSafeEval(geometry: Geometry): Any = {
+    geometry match {
+      case point: Point => point.getCoordinate.getZ
+      case _ => null
+    }
+  }
 
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override def dataType: DataType = DoubleType
+
+  override def children: Seq[Expression] = inputExpressions
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
+}
+
+case class ST_StartPoint(inputExpressions: Seq[Expression])
+  extends UnaryGeometryExpression with CodegenFallback {
+  assert(inputExpressions.length == 1)
+
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry match {
       case line: LineString => line.getPointN(0).toGenericArrayData
       case _ => null
@@ -820,14 +792,11 @@ case class ST_StartPoint(inputExpressions: Seq[Expression])
 
 
 case class ST_Boundary(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
-  override def nullable: Boolean = true
+  extends UnaryGeometryExpression with CodegenFallback {
 
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     val geometryBoundary = geometry.getBoundary
     geometryBoundary.toGenericArrayData
-
   }
 
   override def dataType: DataType = GeometryUDT
@@ -841,17 +810,11 @@ case class ST_Boundary(inputExpressions: Seq[Expression])
 
 
 case class ST_MinimumBoundingRadius(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   private val geometryFactory = new GeometryFactory()
 
-  override def nullable: Boolean = true
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
-    geometry match {
-      case geom: Geometry => getMinimumBoundingRadius(geom)
-      case _ => null
-    }
+  override protected def nullSafeEval(geometry: Geometry): Any = {
+    getMinimumBoundingRadius(geometry)
   }
 
   private def getMinimumBoundingRadius(geom: Geometry): InternalRow = {
@@ -1008,11 +971,9 @@ case class ST_LineInterpolatePoint(inputExpressions: Seq[Expression])
 
 
 case class ST_EndPoint(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
-  override def nullable: Boolean = true
+  extends UnaryGeometryExpression with CodegenFallback {
 
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry match {
       case string: LineString => string.getEndPoint.toGenericArrayData
       case _ => null
@@ -1030,11 +991,9 @@ case class ST_EndPoint(inputExpressions: Seq[Expression])
 }
 
 case class ST_ExteriorRing(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
-  override def nullable: Boolean = true
+  extends UnaryGeometryExpression with CodegenFallback {
 
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry match {
       case polygon: Polygon => polygon.getExteriorRing.toGenericArrayData
       case _ => null
@@ -1112,13 +1071,10 @@ case class ST_InteriorRingN(inputExpressions: Seq[Expression])
 }
 
 case class ST_Dump(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     val geometryCollection = geometry match {
       case collection: GeometryCollection => {
         val numberOfGeometries = collection.getNumGeometries
@@ -1141,13 +1097,10 @@ case class ST_Dump(inputExpressions: Seq[Expression])
 }
 
 case class ST_DumpPoints(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     ArrayData.toArrayData(geometry.getPoints.map(geom => geom.toGenericArrayData))
   }
 
@@ -1162,13 +1115,10 @@ case class ST_DumpPoints(inputExpressions: Seq[Expression])
 
 
 case class ST_IsClosed(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry match {
       case circle: Circle => true
       case point: MultiPoint => true
@@ -1192,13 +1142,10 @@ case class ST_IsClosed(inputExpressions: Seq[Expression])
 }
 
 case class ST_NumInteriorRings(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry match {
       case polygon: Polygon => polygon.getNumInteriorRing
       case _: Geometry => null
@@ -1299,13 +1246,10 @@ case class ST_RemovePoint(inputExpressions: Seq[Expression])
 }
 
 case class ST_IsRing(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = true
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = inputExpressions.head.toGeometry(input)
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry match {
       case string: LineString => string.isSimple & string.isClosed
       case _ => null
@@ -1330,13 +1274,10 @@ case class ST_IsRing(inputExpressions: Seq[Expression])
   * @param inputExpressions Geometry
   */
 case class ST_NumGeometries(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions.head.eval(input).asInstanceOf[ArrayData])
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     geometry.getNumGeometries()
   }
 
@@ -1356,13 +1297,10 @@ case class ST_NumGeometries(inputExpressions: Seq[Expression])
   * @param inputExpressions Geometry
   */
 case class ST_FlipCoordinates(inputExpressions: Seq[Expression])
-  extends Expression with CodegenFallback {
+  extends UnaryGeometryExpression with CodegenFallback {
   assert(inputExpressions.length == 1)
 
-  override def nullable: Boolean = false
-
-  override def eval(input: InternalRow): Any = {
-    val geometry = GeometrySerializer.deserialize(inputExpressions(0).eval(input).asInstanceOf[ArrayData])
+  override protected def nullSafeEval(geometry: Geometry): Any = {
     GeomUtils.flipCoordinates(geometry)
     geometry.toGenericArrayData
   }
@@ -1506,3 +1444,97 @@ case class ST_GeoHash(inputExpressions: Seq[Expression])
     copy(inputExpressions = newChildren)
   }
 }
+
+/**
+ * Return the difference between geometry A and B
+ *
+ * @param inputExpressions
+ */
+case class ST_Difference(inputExpressions: Seq[Expression])
+  extends BinaryGeometryExpression with CodegenFallback {
+  assert(inputExpressions.length == 2)
+
+  lazy val GeometryFactory = new GeometryFactory()
+  lazy val emptyPolygon = GeometryFactory.createPolygon(null, null)
+
+  override protected def nullSafeEval(leftGeometry: Geometry, rightGeometry: Geometry): Any = {
+    val isIntersects = leftGeometry.intersects(rightGeometry)
+    lazy val isRightContainsLeft = rightGeometry.contains(leftGeometry)
+
+    if (!isIntersects) {
+      new GenericArrayData(GeometrySerializer.serialize(leftGeometry))
+    }
+
+    if (isIntersects && isRightContainsLeft) {
+      new GenericArrayData(GeometrySerializer.serialize(emptyPolygon))
+    }
+
+    new GenericArrayData(GeometrySerializer.serialize(leftGeometry.difference(rightGeometry)))
+  }
+
+  override def dataType: DataType = GeometryUDT
+
+  override def children: Seq[Expression] = inputExpressions
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
+}
+
+/**
+ * Return the symmetrical difference between geometry A and B
+ *
+ * @param inputExpressions
+ */
+case class ST_SymDifference(inputExpressions: Seq[Expression])
+  extends BinaryGeometryExpression with CodegenFallback {
+  assert(inputExpressions.length == 2)
+
+  override protected def nullSafeEval(leftGeometry: Geometry, rightGeometry: Geometry): Any = {
+    new GenericArrayData(GeometrySerializer.serialize(leftGeometry.symDifference(rightGeometry)))
+  }
+
+  override def dataType: DataType = GeometryUDT
+
+  override def children: Seq[Expression] = inputExpressions
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
+}
+
+/**
+ * Return the union of geometry A and B
+ *
+ * @param inputExpressions
+ */
+case class ST_Union(inputExpressions: Seq[Expression])
+  extends BinaryGeometryExpression with CodegenFallback {
+  assert(inputExpressions.length == 2)
+
+  override protected def nullSafeEval(leftGeometry: Geometry, rightGeometry: Geometry): Any = {
+    new GenericArrayData(GeometrySerializer.serialize(leftGeometry.union(rightGeometry)))
+  }
+
+  override def dataType: DataType = GeometryUDT
+
+  override def children: Seq[Expression] = inputExpressions
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
+}
+
+case class ST_Multi(inputExpressions: Seq[Expression]) extends UnaryGeometryExpression with CodegenFallback{
+  override def dataType: DataType = GeometryUDT
+
+  override def children: Seq[Expression] = inputExpressions
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
+
+  override protected def nullSafeEval(geometry: Geometry): Any ={
+    Collect.createMultiGeometry(Seq(geometry)).toGenericArrayData
+  }
+}
\ No newline at end of file
