@@ -52,10 +52,14 @@ case class ST_Contains(inputExpressions: Seq[Expression])
 
     val rightGeometry = GeometrySerializer.deserialize(rightArray)
 
-    return leftGeometry.covers(rightGeometry)
+    leftGeometry.covers(rightGeometry)
   }
 
   override def dataType = BooleanType
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
 }
 
 /**
@@ -82,10 +86,14 @@ case class ST_Intersects(inputExpressions: Seq[Expression])
 
     val rightGeometry = GeometrySerializer.deserialize(rightArray)
 
-    return leftGeometry.intersects(rightGeometry)
+    leftGeometry.intersects(rightGeometry)
   }
 
   override def dataType = BooleanType
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
 }
 
 /**
@@ -112,10 +120,14 @@ case class ST_Within(inputExpressions: Seq[Expression])
 
     val rightGeometry = GeometrySerializer.deserialize(rightArray)
 
-    return leftGeometry.coveredBy(rightGeometry)
+    leftGeometry.coveredBy(rightGeometry)
   }
 
   override def dataType = BooleanType
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
 }
 
 
@@ -126,6 +138,8 @@ case class ST_Within(inputExpressions: Seq[Expression])
   */
 case class ST_Crosses(inputExpressions: Seq[Expression])
   extends ST_Predicate with CodegenFallback {
+  assert(inputExpressions.length == 2)
+
   override def nullable: Boolean = false
 
   override def toString: String = s" **${ST_Crosses.getClass.getName}**  "
@@ -133,8 +147,6 @@ case class ST_Crosses(inputExpressions: Seq[Expression])
   override def children: Seq[Expression] = inputExpressions
 
   override def eval(inputRow: InternalRow): Any = {
-    assert(inputExpressions.length == 2)
-
     val leftArray = inputExpressions(0).eval(inputRow).asInstanceOf[ArrayData]
     val rightArray = inputExpressions(1).eval(inputRow).asInstanceOf[ArrayData]
 
@@ -142,10 +154,14 @@ case class ST_Crosses(inputExpressions: Seq[Expression])
 
     val rightGeometry = GeometrySerializer.deserialize(rightArray)
 
-    return leftGeometry.crosses(rightGeometry)
+    leftGeometry.crosses(rightGeometry)
   }
 
   override def dataType = BooleanType
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
 }
 
 
@@ -173,10 +189,14 @@ case class ST_Overlaps(inputExpressions: Seq[Expression])
 
     val rightGeometry = GeometrySerializer.deserialize(rightArray)
 
-    return leftGeometry.overlaps(rightGeometry)
+    leftGeometry.overlaps(rightGeometry)
   }
 
   override def dataType = BooleanType
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
 }
 
 /**
@@ -203,10 +223,14 @@ case class ST_Touches(inputExpressions: Seq[Expression])
 
     val rightGeometry = GeometrySerializer.deserialize(rightArray)
 
-    return leftGeometry.touches(rightGeometry)
+    leftGeometry.touches(rightGeometry)
   }
 
   override def dataType = BooleanType
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
 }
 
 /**
@@ -237,10 +261,12 @@ case class ST_Equals(inputExpressions: Seq[Expression])
     // Returns GeometryCollection object
     val symDifference = leftGeometry.symDifference(rightGeometry)
 
-    val isEqual = symDifference.isEmpty
-
-    return isEqual
+    symDifference.isEmpty
   }
 
   override def dataType = BooleanType
+
+  protected def withNewChildrenInternal(newChildren: IndexedSeq[Expression]) = {
+    copy(inputExpressions = newChildren)
+  }
 }
\ No newline at end of file
